#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from datetime import datetime
import threading
from threading import Lock

from .BinaryEncodingCache import BinaryEncodingCache

class Checksum:
    def __init__(self, value: int, expiration: int) -> None:
        self.value = value
        self.expiration = expiration


class ClientBinaryStrategy:

    def __init__(self, binary_encoding_cache: 'BinaryEncodingCache') -> None:
        self.binary_encoding_cache = binary_encoding_cache
        self.primary: Checksum | None = None
        self.secondary: Checksum | None = None
        self.last_update = datetime.now()
        self.lock = Lock()

    def update_checksum(self, new_checksum: int) -> None:
        with self.lock:
            if self.primary is None:
                self.primary = Checksum(new_checksum, 0)
                return

            if self.primary.value != new_checksum:
                expired_checksum = self.secondary
                self.secondary = self.primary
                self.primary = Checksum(new_checksum, 0)
                self.secondary.expiration += 1

                if expired_checksum:
                    self.binary_encoding_cache.remove(expired_checksum.value)

                return

            self.last_update = datetime.now()

    def get_current_checksums(self) -> list[int]:
        with self.lock:
            if self.primary is None:
                return []
            elif self.secondary is None:
                return [self.primary.value]
            else:
                minutes_since_last_update = (
                    datetime.now() - self.last_update).total_seconds() / 60

                # Every 10 minute interval of non-use is a penalty point
                penalty = int(minutes_since_last_update // 10) + 1

                self.secondary.expiration += 1 * penalty

                if self.secondary.expiration > 5:
                    self.secondary = None
                    return [self.primary.value]
                else:
                    return [self.primary.value, self.secondary.value]



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import re
from typing import TYPE_CHECKING, cast


if TYPE_CHECKING:
    from ..types.TError import TError
    from ..types.TStruct import TStruct
    from ..types.TType import TType


def apply_error_to_parsed_types(error: 'TError', 
                                parsed_types: dict[str, 'TType'],
                                schema_keys_to_document_names: dict[str, str],
                                schema_keys_to_index: dict[str, int],
                                document_names_to_json: dict[str, str],
                                fn_error_regexes: dict[str, str]) -> None:
    from ...internal.schema.SchemaParseFailure import SchemaParseFailure
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ...internal.schema.DocumentLocators import resolve_document_coordinates
    from ..types.TUnion import TUnion

    parse_failures = []

    error_key = error.name
    error_index = schema_keys_to_index[error_key]
    document_name = schema_keys_to_document_names[error_key]

    for parsed_type_name in parsed_types.keys():
        if not parsed_type_name.startswith('fn.') or parsed_type_name.endswith('.->'):
            continue
        parsed_type = parsed_types[parsed_type_name + '.->']
        f = cast(TUnion, parsed_type)

        fn_name = parsed_type_name
        fn_error_regex = fn_error_regexes[fn_name]

        regex = re.compile(fn_error_regex)

        fn_result = f
        fn_result_tags = fn_result.tags
        error_errors = error.errors
        error_tags = error_errors.tags

        matcher = regex.match(error_key)
        if not matcher:
            continue

        for error_tag_name, error_tag in error_tags.items():
            new_key = error_tag_name

            if new_key in fn_result_tags:
                other_path_index = schema_keys_to_index[fn_name]
                error_tag_index = error.errors.tag_indices[new_key]
                other_document_name = schema_keys_to_document_names[fn_name]
                fn_error_tag_index = f.tag_indices[new_key]
                other_final_path = [other_path_index,
                                    "->", fn_error_tag_index, new_key]
                other_location_pseudo_json = resolve_document_coordinates(
                    other_final_path, other_document_name, document_names_to_json)
                parse_failures.append(SchemaParseFailure(
                    document_name,
                    [error_index, error_key,
                     error_tag_index, new_key],
                    "PathCollision",
                    {"document": other_document_name,
                     "path": other_final_path,
                        "location": other_location_pseudo_json}
                ))
            fn_result_tags[new_key] = error_tag

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, document_names_to_json)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

def partially_matches(whole_list: list[object], part_element: object) -> bool:
    from ...internal.mock.IsSubMapEntryEqual import is_sub_map_entry_equal

    for whole_element in whole_list:
        if is_sub_map_entry_equal(part_element, whole_element):
            return True
    return False



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING
from ...Message import Message

if TYPE_CHECKING:
    from ..types.TUnion import TUnion
    from ...internal.validation.ValidationFailure import ValidationFailure


def get_invalid_error_message(error: str, validation_failures: list['ValidationFailure'],
                              result_union_type: 'TUnion', response_headers: dict[str, object]) -> 'Message':
    from ...internal.validation.MapValidationFailuresToInvalidFieldCases import map_validation_failures_to_invalid_field_cases
    from ...internal.validation.ValidateResult import validate_result

    validation_failure_cases = map_validation_failures_to_invalid_field_cases(
        validation_failures)
    new_error_result: dict[str, object] = {
        error: {
            "cases": validation_failure_cases
        }
    }

    validate_result(result_union_type, new_error_result)
    return Message(response_headers, new_error_result)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import cast, TYPE_CHECKING

from ..Message import Message
from ..internal.validation.InvalidMessage import InvalidMessage
from ..internal.validation.InvalidMessageBody import InvalidMessageBody

if TYPE_CHECKING:
    from ..Serialization import Serialization
    from ..internal.binary.BinaryEncoder import BinaryEncoder
    from ..internal.binary.Base64Encoder import Base64Encoder


def deserialize_internal(message_bytes: bytes, serializer: 'Serialization',
                         binary_encoder: 'BinaryEncoder',
                         base64_encoder: 'Base64Encoder') -> 'Message':
    message_as_pseudo_json: object
    is_msg_pack: bool

    try:
        if message_bytes[0] == 0x92:  # MsgPack
            is_msg_pack = True
            message_as_pseudo_json = serializer.from_msgpack(message_bytes)
        else:
            is_msg_pack = False
            message_as_pseudo_json = serializer.from_json(message_bytes)
    except Exception as e:
        raise InvalidMessage() from e

    if not isinstance(message_as_pseudo_json, list):
        raise InvalidMessage()

    message_as_pseudo_json_list = cast(list[object], message_as_pseudo_json)

    if len(message_as_pseudo_json_list) != 2:
        raise InvalidMessage()

    final_message_as_pseudo_json_list: list[object]
    if is_msg_pack:
        final_message_as_pseudo_json_list = binary_encoder.decode(
            message_as_pseudo_json_list)
    else:
        final_message_as_pseudo_json_list = base64_encoder.decode(message_as_pseudo_json_list)

    if not isinstance(final_message_as_pseudo_json_list[0], dict):
        raise InvalidMessage()

    headers = cast(dict[str, object], final_message_as_pseudo_json_list[0])

    if not isinstance(final_message_as_pseudo_json_list[1], dict):
        raise InvalidMessage()

    body = cast(dict[str, object], final_message_as_pseudo_json_list[1])

    if len(body) != 1:
        raise InvalidMessageBody()

    if not isinstance(next(iter(body.values())), dict):
        raise InvalidMessageBody()

    return Message(headers, body)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Dict, List, Set, cast


def catch_header_collisions(
    telepact_schema_name_to_pseudo_json: Dict[str, List[object]],
    header_keys: Set[str],
    keys_to_index: Dict[str, int],
    schema_keys_to_document_names: Dict[str, str],
    document_names_to_json: Dict[str, str],
) -> None:
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ...internal.schema.SchemaParseFailure import SchemaParseFailure
    from ...internal.schema.DocumentLocators import resolve_document_coordinates

    parse_failures = []

    header_keys_list = list(header_keys)
    header_keys_list.sort(key=lambda k: (
        schema_keys_to_document_names[k], keys_to_index[k]))

    for i in range(len(header_keys_list)):
        for j in range(i + 1, len(header_keys_list)):
            def_key = header_keys_list[i]
            other_def_key = header_keys_list[j]

            index = keys_to_index[def_key]
            other_index = keys_to_index[other_def_key]

            document_name = schema_keys_to_document_names[def_key]
            other_document_name = schema_keys_to_document_names[other_def_key]

            telepact_schema_pseudo_json = telepact_schema_name_to_pseudo_json[document_name]
            other_telepact_schema_pseudo_json = telepact_schema_name_to_pseudo_json[
                other_document_name]

            def_ = cast(dict[str, object], telepact_schema_pseudo_json[index])
            other_def = cast(dict[str, object],
                             other_telepact_schema_pseudo_json[other_index])

            header_def = cast(dict[str, object], def_[def_key])
            other_header_def = cast(
                dict[str, object], other_def[other_def_key])

            header_collisions = [
                k for k in header_def if k in other_header_def]
            for header_collision in header_collisions:
                this_path = [index, def_key, header_collision]
                this_location = resolve_document_coordinates(
                    this_path, document_name, document_names_to_json)
                parse_failures.append(
                    SchemaParseFailure(
                        other_document_name,
                        [other_index, other_def_key, header_collision],
                        'PathCollision',
                        {'document': document_name, 'path': this_path,
                            'location': this_location},
                    )
                )

            res_header_def = cast(dict[str, object], def_['->'])
            other_res_header_def = cast(dict[str, object], other_def['->'])

            res_header_collisions = [
                k for k in res_header_def if k in other_res_header_def]
            for res_header_collision in res_header_collisions:
                this_path = [index, '->', res_header_collision]
                this_location = resolve_document_coordinates(
                    this_path, document_name, document_names_to_json)
                parse_failures.append(
                    SchemaParseFailure(
                        other_document_name,
                        [other_index, '->', res_header_collision],
                        'PathCollision',
                        {'document': document_name, 'path': this_path,
                            'location': this_location},
                    )
                )

    if parse_failures:
        raise TelepactSchemaParseError(parse_failures, document_names_to_json)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import base64
from typing import TYPE_CHECKING, cast

from ..RandomGenerator import RandomGenerator
from .generation.GenerateContext import GenerateContext

if TYPE_CHECKING:
    from ..TelepactSchema import TelepactSchema
    from .types.TFieldDeclaration import TFieldDeclaration
    from .types.TType import TType


_EXAMPLE_COLLECTION_LENGTH = 2


def get_api_definitions_with_examples(
    telepact_schema: 'TelepactSchema',
    include_internal: bool,
) -> list[object]:
    definitions = telepact_schema.full if include_internal else telepact_schema.original
    default_fn_scope = _get_default_fn_scope(telepact_schema.parsed)

    return [
        _add_examples_to_definition(
            cast(dict[str, object], definition),
            telepact_schema,
            default_fn_scope,
        )
        for definition in definitions
    ]


def _add_examples_to_definition(
    definition: dict[str, object],
    telepact_schema: 'TelepactSchema',
    default_fn_scope: str,
) -> dict[str, object]:
    schema_key = _get_schema_key(definition)
    cloned_definition = dict(definition)

    if schema_key.startswith('info.'):
        cloned_definition['example'] = {}
        return cloned_definition

    random_generator = RandomGenerator(_EXAMPLE_COLLECTION_LENGTH, _EXAMPLE_COLLECTION_LENGTH)

    if schema_key.startswith('fn.'):
        ctx = GenerateContext(True, False, True, schema_key, random_generator)
        cloned_definition['inputExample'] = _normalize_example_value(
            telepact_schema.parsed[schema_key].generate_random_value(
                None, False, [], ctx)
        )
        cloned_definition['outputExample'] = _normalize_example_value(
            telepact_schema.parsed[
                schema_key + '.->'
            ].generate_random_value(None, False, [], ctx)
        )
        return cloned_definition

    if schema_key.startswith('headers.'):
        ctx = GenerateContext(True, False, True, default_fn_scope, random_generator)
        cloned_definition['inputExample'] = _generate_header_example(
            cast(dict[str, object], definition.get(schema_key, {})),
            telepact_schema.parsed_request_headers,
            ctx,
        )
        cloned_definition['outputExample'] = _generate_header_example(
            cast(dict[str, object], definition.get('->', {})),
            telepact_schema.parsed_response_headers,
            ctx,
        )
        return cloned_definition

    if schema_key.startswith('errors.'):
        ctx = GenerateContext(True, False, True, default_fn_scope, random_generator)
        cloned_definition['example'] = _generate_raw_union_example(
            cast(list[object], definition[schema_key]),
            telepact_schema,
            ctx,
        )
        return cloned_definition

    ctx = GenerateContext(True, False, True, default_fn_scope, random_generator)
    cloned_definition['example'] = _normalize_example_value(
        telepact_schema.parsed[schema_key].generate_random_value(
            None, False, [], ctx)
    )
    return cloned_definition


def _generate_header_example(
    header_definition: dict[str, object],
    parsed_headers: dict[str, 'TFieldDeclaration'],
    ctx: GenerateContext,
) -> dict[str, object]:
    example: dict[str, object] = {}

    for header_name in sorted(header_definition.keys()):
        example[header_name] = _normalize_example_value(
            parsed_headers[header_name].type_declaration.generate_random_value(
                None, False, ctx)
        )

    return example


def _generate_raw_union_example(
    union_definition: list[object],
    telepact_schema: 'TelepactSchema',
    ctx: GenerateContext,
) -> dict[str, object]:
    tags = sorted([
        cast(tuple[str, dict[str, object]], next(
            (key, cast(dict[str, object], value))
            for key, value in cast(dict[str, object], tag_definition).items()
            if key != '///'
        ))
        for tag_definition in union_definition
    ], key=lambda entry: entry[0])
    tag_name, tag_payload = tags[ctx.random_generator.next_int_with_ceiling(len(tags))]
    return {
        tag_name: _generate_raw_struct_example(tag_payload, telepact_schema, ctx),
    }


def _generate_raw_struct_example(
    struct_definition: dict[str, object],
    telepact_schema: 'TelepactSchema',
    ctx: GenerateContext,
) -> dict[str, object]:
    example: dict[str, object] = {}

    for field_name in sorted(struct_definition.keys()):
        optional = field_name.endswith('!')
        if optional:
            if not ctx.include_optional_fields or (
                ctx.randomize_optional_fields and ctx.random_generator.next_boolean()
            ):
                continue
        elif not ctx.always_include_required_fields and ctx.random_generator.next_boolean():
            continue

        example[field_name] = _generate_raw_type_example(
            struct_definition[field_name],
            telepact_schema,
            ctx,
        )

    return example


def _generate_raw_type_example(
    type_expression: object,
    telepact_schema: 'TelepactSchema',
    ctx: GenerateContext,
) -> object:
    if isinstance(type_expression, str):
        nullable = type_expression.endswith('?')
        non_nullable_type_expression = type_expression[:-1] if nullable else type_expression
        if nullable and ctx.random_generator.next_boolean():
            return None

        if non_nullable_type_expression == 'boolean':
            return ctx.random_generator.next_boolean()
        if non_nullable_type_expression == 'integer':
            return ctx.random_generator.next_int()
        if non_nullable_type_expression == 'number':
            return ctx.random_generator.next_double()
        if non_nullable_type_expression == 'string':
            return ctx.random_generator.next_string()
        if non_nullable_type_expression == 'any':
            select_type = ctx.random_generator.next_int_with_ceiling(3)
            if select_type == 0:
                return ctx.random_generator.next_boolean()
            if select_type == 1:
                return ctx.random_generator.next_int()
            return ctx.random_generator.next_string()
        if non_nullable_type_expression == 'bytes':
            return base64.b64encode(ctx.random_generator.next_bytes()).decode()

        return _normalize_example_value(
            telepact_schema.parsed[non_nullable_type_expression].generate_random_value(
                None, False, [], ctx)
        )

    if isinstance(type_expression, list):
        return [
            _generate_raw_type_example(type_expression[0], telepact_schema, ctx)
            for _ in range(ctx.random_generator.next_collection_length())
        ]

    if isinstance(type_expression, dict):
        if set(type_expression.keys()) == {'string'}:
            return {
                ctx.random_generator.next_string(): _generate_raw_type_example(
                    type_expression['string'], telepact_schema, ctx)
                for _ in range(ctx.random_generator.next_collection_length())
            }

    return None


def _normalize_example_value(value: object) -> object:
    if isinstance(value, bytes):
        return base64.b64encode(value).decode()

    if isinstance(value, list):
        return [_normalize_example_value(entry) for entry in value]

    if isinstance(value, dict):
        return {
            key: _normalize_example_value(entry)
            for key, entry in value.items()
        }

    return value


def _get_default_fn_scope(parsed_types: dict[str, 'TType']) -> str:
    non_internal_functions = sorted([
        schema_key for schema_key in parsed_types.keys()
        if schema_key.startswith('fn.') and not schema_key.endswith('.->') and not schema_key.endswith('_')
    ])
    if non_internal_functions:
        return non_internal_functions[0]

    all_functions = sorted([
        schema_key for schema_key in parsed_types.keys()
        if schema_key.startswith('fn.') and not schema_key.endswith('.->')
    ])
    if all_functions:
        return all_functions[0]

    return 'fn.ping_'


def _get_schema_key(definition: dict[str, object]) -> str:
    for key in definition.keys():
        if key not in {'///', '->', '_errors'}:
            return key

    raise ValueError(f'Schema entry has no schema key: {definition}')



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Dict, List, Union

from ...internal.generation.GenerateContext import GenerateContext


def generate_random_select(possible_selects: dict[str, object], ctx: 'GenerateContext') -> object:
    possible_select = possible_selects[ctx.fn_scope]
    return sub_select(possible_select, ctx)


def sub_select(possible_select_section: object, ctx: 'GenerateContext') -> object:
    if isinstance(possible_select_section, list):
        selected_field_names = []

        for field_name in possible_select_section:
            if ctx.random_generator.next_boolean():
                selected_field_names.append(field_name)

        return sorted(selected_field_names)
    elif isinstance(possible_select_section, dict):
        selected_section: dict[str, object] = {}

        keys = sorted(possible_select_section.keys())

        for key in keys:
            value = possible_select_section[key]
            if ctx.random_generator.next_boolean():
                result = sub_select(value, ctx)
                if isinstance(result, dict) and not result:
                    continue
                selected_section[key] = result

        return selected_section
    else:
        raise ValueError('Invalid possible_select_section')



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import inspect
import math
from ctypes import c_int32


def _find_stack() -> str:
    stack = inspect.stack()
    for i in range(1, len(stack)):
        stack_str = f'{stack[i]}'
        if 'RandomGenerator' not in stack_str:
            return f'{stack[i].function} {stack[i-1].function}'
    return 'unknown'

words = [
    "alpha", "beta", "gamma", "delta", "epsilon", "zeta", "eta", "theta",
    "iota", "kappa", "lambda", "mu", "nu", "xi", "omicron", "pi",
    "rho", "sigma", "tau", "upsilon", "phi", "chi", "psi", "omega"
]

class RandomGenerator:
    def __init__(self, collection_length_min: int, collection_length_max: int):
        self.set_seed(0)
        self.collection_length_min = collection_length_min
        self.collection_length_max = collection_length_max
        self.count = 0

    def set_seed(self, seed: int) -> None:
        self.seed = c_int32(1) if seed == 0 else c_int32(seed)

    def next_int(self) -> int:
        x: c_int32 = c_int32(self.seed.value)
        x = c_int32(x.value ^ (x.value << 16))
        x = c_int32(x.value ^ (x.value >> 11))
        x = c_int32(x.value ^ (x.value << 5))
        self.seed = c_int32(1) if x.value == 0 else c_int32(x.value)
        self.count += 1
        result = c_int32(self.seed.value & 0x7fffffff)
        # print(f'{self.count} {result} {_find_stack()}')
        return result.value

    def next_int_with_ceiling(self, ceiling: int) -> int:
        if ceiling == 0:
            return 0
        return self.next_int() % ceiling

    def next_boolean(self) -> bool:
        return self.next_int_with_ceiling(31) > 15
    
    def next_bytes(self) -> bytes:
        import struct
        return struct.pack(">i", self.next_int())

    def next_string(self) -> str:
        index = self.next_int_with_ceiling(len(words))
        return words[index]

    def next_double(self) -> float:
        return float(self.next_int() & 0x7fffffff) / float(0x7fffffff)

    def next_collection_length(self) -> int:
        return self.next_int_with_ceiling(self.collection_length_max - self.collection_length_min) + self.collection_length_min



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import json
import math
from typing import Callable

from .BuildDocumentLocatorFromYamlAst import create_document_locator_from_yaml_text


def _normalize_json_compatible_value(value: object) -> object:
    if value is None or isinstance(value, (str, bool, int)):
        return value

    if isinstance(value, float):
        if not math.isfinite(value):
            raise ValueError('YAML values must be JSON-compatible')
        return value

    if isinstance(value, list):
        return [_normalize_json_compatible_value(child) for child in value]

    if isinstance(value, dict):
        normalized: dict[str, object] = {}
        for key, child in value.items():
            if not isinstance(key, str):
                raise ValueError('YAML values must be JSON-compatible')
            normalized[key] = _normalize_json_compatible_value(child)
        return normalized

    raise ValueError('YAML values must be JSON-compatible')


def parse_telepact_yaml(text: str) -> tuple[str, Callable[[list[object]], dict[str, object]] | None]:
    import yaml

    try:
        parsed = yaml.safe_load(text)
    except yaml.YAMLError as e:
        raise ValueError(str(e)) from e

    normalized = _normalize_json_compatible_value(parsed)
    if not isinstance(normalized, list):
        raise ValueError('Telepact YAML root must be a sequence')

    return json.dumps(normalized, separators=(',', ':')), create_document_locator_from_yaml_text(text)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast

from ...Message import Message
from ...internal.mock.MockInvocation import MockInvocation
from ...internal.mock.MockStub import MockStub

if TYPE_CHECKING:
    from ...RandomGenerator import RandomGenerator
    from ...TelepactSchema import TelepactSchema


async def mock_handle(request_message: 'Message', stubs: list['MockStub'], invocations: list['MockInvocation'],
                      random: 'RandomGenerator', telepact_schema: 'TelepactSchema', enable_generated_default_stub: bool,
                      enable_optional_field_generation: bool, randomize_optional_field_generation: bool) -> 'Message':
    from ...internal.mock.IsSubMap import is_sub_map
    from ...internal.mock.Verify import verify
    from ...internal.mock.VerifyNoMoreInteractions import verify_no_more_interactions
    from ...TelepactError import TelepactError
    from ..types.TUnion import TUnion
    from ...internal.generation.GenerateContext import GenerateContext

    header: dict[str, object] = request_message.headers

    enable_generation_stub = header.get("@gen_", False)
    function_name = request_message.get_body_target()
    argument = request_message.get_body_payload()

    if function_name == "fn.createStub_":
        given_stub = cast(dict[str, object], argument.get("stub"))

        stub_call = next((e for e in given_stub.items()
                         if e[0].startswith("fn.")))
        stub_function_name, stub_arg = cast(
            tuple[str, dict[str, object]], stub_call)
        stub_result = cast(dict[str, object], given_stub.get("->"))
        allow_argument_partial_match = not argument.get("strictMatch!", False)
        stub_count = cast(int, argument.get("count!", -1))

        stub = MockStub(stub_function_name, stub_arg,
                        stub_result, allow_argument_partial_match, stub_count)

        stubs.insert(0, stub)
        return Message({}, {"Ok_": {}})
    elif function_name == "fn.verify_":
        given_call = cast(dict[str, object], argument.get("call"))

        call = next((e for e in given_call.items()
                    if e[0].startswith("fn.")), None)
        call_function_name, call_arg = cast(
            tuple[str, dict[str, object]], call)
        verify_times = cast(dict[str, object], argument.get(
            "count!", {"AtLeast": {"times": 1}}))
        strict_match = cast(bool, argument.get("strictMatch!", False))

        verification_result = verify(
            call_function_name, call_arg, strict_match, verify_times, invocations)
        return Message({}, verification_result)
    elif function_name == "fn.verifyNoMoreInteractions_":
        verification_result = verify_no_more_interactions(
            invocations)
        return Message({}, verification_result)
    elif function_name == "fn.clearCalls_":
        invocations.clear()
        return Message({}, {"Ok_": {}})
    elif function_name == "fn.clearStubs_":
        stubs.clear()
        return Message({}, {"Ok_": {}})
    elif function_name == "fn.setRandomSeed_":
        given_seed = cast(int, argument.get("seed"))

        random.set_seed(given_seed)
        return Message({}, {"Ok_": {}})
    else:
        invocations.append(MockInvocation(function_name, dict(argument)))

        definition = cast(TUnion, telepact_schema.parsed.get(function_name + '.->'))

        for stub in stubs:
            if stub.count == 0:
                continue
            if stub.when_function == function_name:
                if stub.allow_argument_partial_match:
                    if is_sub_map(stub.when_argument, argument):
                        use_blueprint_value = True
                        include_optional_fields = False
                        always_include_required_fields = True
                        result_init = definition.generate_random_value(stub.then_result, use_blueprint_value, [],
                                                                              GenerateContext(
                                                                              include_optional_fields,
                                                                              randomize_optional_field_generation,
                                                                              always_include_required_fields,
                                                                              function_name, random))
                        result = cast(dict[str, object], result_init)
                        if stub.count > 0:
                            stub.count -= 1
                        return Message({}, result)
                else:
                    if stub.when_argument == argument:
                        use_blueprint_value = True
                        include_optional_fields = False
                        always_include_required_fields = True
                        result_init = definition.generate_random_value(stub.then_result, use_blueprint_value, [],
                                                                              GenerateContext(
                                                                              include_optional_fields,
                                                                              randomize_optional_field_generation,
                                                                              always_include_required_fields,
                                                                              function_name, random))
                        result = cast(dict[str, object], result_init)
                        if stub.count > 0:
                            stub.count -= 1
                        return Message({}, result)

        if not enable_generated_default_stub and not enable_generation_stub:
            return Message({}, {"ErrorNoMatchingStub_": {}})

        if definition is not None:
            result_union = definition
            ok_struct_ref = result_union.tags["Ok_"]
            use_blueprint_value = True
            include_optional_fields = True
            always_include_required_fields = True
            random_ok_struct_init = ok_struct_ref.generate_random_value({}, use_blueprint_value, [],
                                                                        GenerateContext(
                                                                        include_optional_fields,
                                                                        randomize_optional_field_generation,
                                                                        always_include_required_fields,
                                                                        function_name, random))
            random_ok_struct = cast(dict[str, object], random_ok_struct_init)
            return Message({}, {"Ok_": random_ok_struct})
        else:
            raise TelepactError(
                "Unexpected unknown function: {}".format(function_name))



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import cast

from .GenerateContext import GenerateContext
from .GenerateRandomStruct import generate_random_struct
from ..types.TType import TType


def generate_random_mock_stub(types: dict[str, TType], ctx: GenerateContext) -> object:
    from ..types.TUnion import TUnion

    functions = [key for key in types.keys() if key.startswith('fn.') and not key.endswith('.->') and not key.endswith('_')]

    functions.sort()

    index = ctx.random_generator.next_int_with_ceiling(len(functions))

    selected_fn_name = functions[index]
    selected_fn = cast(TUnion, types[selected_fn_name])
    selected_result = cast(TUnion, types[selected_fn_name + '.->'])

    arg_fields = selected_fn.tags[selected_fn_name].fields
    ok_fields = selected_result.tags['Ok_'].fields

    arg = generate_random_struct(None, False, arg_fields, ctx.copy(
        always_include_required_fields=False))
    ok_result = generate_random_struct(None, False,
                                       ok_fields, ctx.copy(always_include_required_fields=False))

    return {
        selected_fn_name: arg,
        '->': {
            'Ok_': ok_result,
        },
    }



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from ...RandomGenerator import RandomGenerator
    from ..types.TTypeDeclaration import TTypeDeclaration
    from ...internal.generation.GenerateContext import GenerateContext


def generate_random_object(blueprint_value: object, use_blueprint_value: bool, type_parameters: list['TTypeDeclaration'], ctx: 'GenerateContext') -> object:
    nested_type_declaration = type_parameters[0]

    if use_blueprint_value:
        starting_obj = cast(dict[str, object], blueprint_value)

        obj = {}
        for key, starting_obj_value in starting_obj.items():
            value = nested_type_declaration.generate_random_value(
                starting_obj_value, True, ctx)
            obj[key] = value

        return obj
    else:
        length = ctx.random_generator.next_collection_length()

        obj = {}
        for i in range(length):
            key = ctx.random_generator.next_string()
            value = nested_type_declaration.generate_random_value(
                None, False, ctx)
            obj[key] = value

        return obj



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
import base64


def server_base64_decode(body: dict[str, object], bytes_paths: dict[str, object]) -> None:
    travel_base64_decode(body, bytes_paths)

def travel_base64_decode(value: object, bytes_paths: object) -> object:
    if isinstance(bytes_paths, dict):
        for key, val in bytes_paths.items():
            if val is True:
                if key == "*" and isinstance(value, list):
                    for i, v in enumerate(value):
                        nv = travel_base64_decode(v, val)
                        value[i] = nv
                elif key == "*" and isinstance(value, dict):
                    for k, v in value.items():
                        nv = travel_base64_decode(v, val)
                        value[k] = nv
                elif isinstance(value, dict):
                    nv = travel_base64_decode(value[key], val)
                    value[key] = nv
                else:
                    raise ValueError(f"Invalid bytes path: {key} for value: {value}")
            else:
                if key == "*" and isinstance(value, list):
                    for i, v in enumerate(value):
                        travel_base64_decode(v, val)
                elif key == "*" and isinstance(value, dict):
                    for k, v in value.items():
                        travel_base64_decode(v, val)
                elif isinstance(value, dict):
                    travel_base64_decode(value[key], val)
                else:
                    raise ValueError(f"Invalid bytes path: {key} for value: {value}")
        return None
    elif bytes_paths is True and (type(value) is str or value is None):
        if value is None:
            return None
        decoded_value = base64.b64decode(value.encode("utf-8"))
        return decoded_value
    else:
        raise ValueError(f"Invalid bytes path: {bytes_paths} for value: {value} of type {type(value)}")


#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Callable, TYPE_CHECKING, Awaitable, Any
from concurrent.futures import Future

from .internal.binary.DefaultBinaryEncodingCache import DefaultBinaryEncodingCache
from .DefaultSerialization import DefaultSerialization
from .Serializer import Serializer
from .internal.binary.ClientBinaryEncoder import ClientBinaryEncoder
from .internal.binary.ClientBase64Encoder import ClientBase64Encoder

if TYPE_CHECKING:
    from .Message import Message


class Client:
    class Options:
        def __init__(self) -> None:
            self.use_binary = False
            self.always_send_json = True
            self.timeout_ms_default = 5000
            self.serialization_impl = DefaultSerialization()

    def __init__(self, adapter: Callable[['Message', 'Serializer'], Awaitable['Message']], options: 'Options'):
        self.adapter = adapter
        self.use_binary_default = options.use_binary
        self.always_send_json = options.always_send_json
        self.timeout_ms_default = options.timeout_ms_default

        binary_encoding_cache = DefaultBinaryEncodingCache()
        binary_encoder = ClientBinaryEncoder(binary_encoding_cache)
        base64_encoder = ClientBase64Encoder()

        self.serializer = Serializer(options.serialization_impl, binary_encoder, base64_encoder)

    async def request(self, request_message: 'Message') -> 'Message':
        from .internal.ClientHandleMessage import client_handle_message
        return await client_handle_message(request_message, self.adapter, self.serializer, self.timeout_ms_default, self.use_binary_default, self.always_send_json)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ...internal.schema.SchemaParseFailure import SchemaParseFailure


def map_schema_parse_failures_to_pseudo_json(schema_parse_failures: list['SchemaParseFailure'], telepact_document_name_to_json: dict[str, str]) -> list[dict[str, object]]:
    from ...internal.schema.DocumentLocators import resolve_document_coordinates
    pseudo_json_list = []
    for f in schema_parse_failures:
        location = resolve_document_coordinates(
            f.path, f.document_name, telepact_document_name_to_json)
        pseudo_json: dict[str, object] = {}
        pseudo_json["document"] = f.document_name
        pseudo_json["location"] = location
        pseudo_json["path"] = f.path
        pseudo_json["reason"] = {f.reason: f.data}
        pseudo_json_list.append(dict(pseudo_json))
    return pseudo_json_list



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
from ...internal.validation.ValidationFailure import ValidationFailure

if TYPE_CHECKING:
    from ...internal.validation.ValidateContext import ValidateContext
    from ..types.TStruct import TStruct
    from ..types.TTypeDeclaration import TTypeDeclaration


def validate_union_tags(reference_tags: dict[str, 'TStruct'], selected_tags: dict[str, object],
                        actual: dict[object, object], ctx: 'ValidateContext') -> list['ValidationFailure']:
    from ...internal.validation.GetTypeUnexpectedValidationFailure import get_type_unexpected_validation_failure
    from ...internal.validation.ValidateUnionStruct import validate_union_struct

    if len(actual) != 1:
        return [
            ValidationFailure([], "ObjectSizeUnexpected", {
                              "actual": len(actual), "expected": 1})
        ]

    union_target, union_payload = cast(
        tuple[str, object], next(iter(actual.items())))

    reference_struct = reference_tags.get(union_target)
    if reference_struct is None:
        return [
            ValidationFailure([union_target], "ObjectKeyDisallowed", {})
        ]

    if isinstance(union_payload, dict):
        nested_validation_failures = validate_union_struct(reference_struct, union_target,
                                                           union_payload, selected_tags, ctx)

        nested_validation_failures_with_path = []
        for failure in nested_validation_failures:
            this_path = [union_target] + failure.path
            nested_validation_failures_with_path.append(
                ValidationFailure(this_path, failure.reason, failure.data)
            )

        return nested_validation_failures_with_path
    else:
        return get_type_unexpected_validation_failure([union_target], union_payload, "Object")



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...internal.mock.MockInvocation import MockInvocation


def verify_no_more_interactions(invocations: list['MockInvocation']) -> dict[str, object]:
    invocations_not_verified = [i for i in invocations if not i.verified]

    if invocations_not_verified:
        unverified_calls_pseudo_json = []
        for invocation in invocations_not_verified:
            unverified_calls_pseudo_json.append(
                {invocation.function_name: invocation.function_argument})
        return {
            "ErrorVerificationFailure": {
                "additionalUnverifiedCalls": unverified_calls_pseudo_json
            }
        }

    return {"Ok_": {}}



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from ...internal.binary.BinaryEncoding import BinaryEncoding


def server_binary_encode(message: list[object], binary_encoder: 'BinaryEncoding') -> list[object]:
    from ...internal.binary.EncodeBody import encode_body
    from ...internal.binary.PackBody import pack_body
    from ...internal.binary.BinaryEncoderUnavailableError import BinaryEncoderUnavailableError

    headers = cast(dict[str, object], message[0])
    message_body = cast(dict[str, object], message[1])
    client_known_binary_checksums = cast(list[int], headers.pop(
        "@clientKnownBinaryChecksums_", None))

    result_tag = list(message_body.keys())[0]

    if result_tag != 'Ok_':
        raise BinaryEncoderUnavailableError()

    if client_known_binary_checksums is None or binary_encoder.checksum not in client_known_binary_checksums:
        headers["@enc_"] = binary_encoder.encode_map

    headers["@bin_"] = [binary_encoder.checksum]
    encoded_message_body = encode_body(message_body, binary_encoder)

    final_encoded_message_body: dict[object, object]
    if headers.get("@pac_") is True:
        final_encoded_message_body = pack_body(encoded_message_body)
    else:
        final_encoded_message_body = encoded_message_body

    return [headers, final_encoded_message_body]



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from .RandomGenerator import RandomGenerator
from .Client import Client
from .Server import Server
from .TestClient import TestClient
from .MockServer import MockServer
from .Message import Message
from .TypedMessage import TypedMessage
from .Response import Response
from .Serializer import Serializer
from .SerializationError import SerializationError
from .Serialization import Serialization
from .TelepactSchema import TelepactSchema
from .MockTelepactSchema import MockTelepactSchema
from .TelepactSchemaParseError import TelepactSchemaParseError
from .TelepactSchemaFiles import TelepactSchemaFiles

__all__ = [
    "RandomGenerator",
    "Client",
    "Server",
    "TestClient",
    "MockServer",
    "Message",
    "Serializer",
    "SerializationError",
    "Serialization",
    "TelepactSchema",
    "MockTelepactSchema",
    "TelepactSchemaParseError",
    "TelepactSchemaFiles",
]



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import asyncio
from typing import Callable, TYPE_CHECKING, cast, Awaitable

if TYPE_CHECKING:
    from ..Message import Message
    from ..Serializer import Serializer


async def client_handle_message(request_message: 'Message',
                                 adapter: Callable[['Message', 'Serializer'], Awaitable['Message']],
                                 serializer: 'Serializer',
                                 timeout_ms_default: int,
                                 use_binary_default: bool,
                                 always_send_json: bool) -> 'Message':
    from ..SerializationError import SerializationError
    from ..TelepactError import TelepactError

    header: dict[str, object] = request_message.headers

    try:
        if "@time_" not in header:
            header["@time_"] = timeout_ms_default

        if use_binary_default:
            header["@binary_"] = True

        if header.get('@binary_', False) and always_send_json:
            header["_forceSendJson"] = True

        timeout_ms = cast(int, header.get("@time_"))

        async with asyncio.timeout(timeout_ms / 1000):
            response_message = await adapter(request_message, serializer)

        if response_message.body == {"ErrorParseFailure_": {"reasons": [{"IncompatibleBinaryEncoding": {}}]}}:
            header["@binary_"] = True
            header["_forceSendJson"] = True

            async with asyncio.timeout(timeout_ms / 1000):
                return await adapter(request_message, serializer)

        return response_message
    except Exception as e:
        if isinstance(e, TelepactError):
            raise
        if isinstance(e, SerializationError):
            raise TelepactError(
                "telepact client serialization or deserialization failed",
                kind="serialization",
                cause=e,
            ) from e
        if isinstance(e, TimeoutError):
            raise TelepactError(
                f"telepact client transport timed out after {timeout_ms}ms",
                kind="transport",
                cause=e,
            ) from e
        raise TelepactError(
            "telepact client transport failed",
            kind="transport",
            cause=e,
        ) from e



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast

from .types.TArray import TArray
from .types.TObject import TObject
from .types.TStruct import TStruct
from .types.TUnion import TUnion

if TYPE_CHECKING:
    from .types.TTypeDeclaration import TTypeDeclaration


def select_struct_fields(type_declaration: 'TTypeDeclaration', value: object,
                         selected_struct_fields: dict[str, object]) -> object:
    type_declaration_type = type_declaration.type
    type_declaration_type_params = type_declaration.type_parameters

    if isinstance(type_declaration_type, TStruct):
        fields = type_declaration_type.fields
        struct_name = type_declaration_type.name
        selected_fields = cast(
            list[str] | None, selected_struct_fields.get(struct_name))
        value_as_map = cast(dict[str, object], value)
        final_map = {}

        for field_name, field_value in value_as_map.items():
            if selected_fields is None or field_name in selected_fields:
                field = fields[field_name]
                field_type_declaration = field.type_declaration
                value_with_selected_fields = select_struct_fields(field_type_declaration, field_value,
                                                                  selected_struct_fields)

                final_map[field_name] = value_with_selected_fields

        return final_map
    elif isinstance(type_declaration_type, TUnion):
        value_as_map = cast(dict[str, object], value)
        union_tag, union_data = cast(
            tuple[str, dict[str, object]], next(iter(value_as_map.items())))

        union_tags = type_declaration_type.tags
        union_struct_reference = union_tags[union_tag]
        union_struct_ref_fields = union_struct_reference.fields
        default_tags_to_fields = {}

        for tag, union_struct in union_tags.items():
            field_names = list(union_struct.fields.keys())
            default_tags_to_fields[tag] = field_names

        union_selected_fields = cast(dict[str, object], selected_struct_fields.get(
            type_declaration_type.name, default_tags_to_fields))
        this_union_tag_selected_fields_default = default_tags_to_fields.get(
            union_tag)
        selected_fields = cast(list[str] | None, union_selected_fields.get(
            union_tag, this_union_tag_selected_fields_default))

        final_map = {}
        for field_name, field_value in union_data.items():
            if selected_fields is None or field_name in selected_fields:
                field = union_struct_ref_fields[field_name]
                value_with_selected_fields = select_struct_fields(field.type_declaration, field_value,
                                                                  selected_struct_fields)

                final_map[field_name] = value_with_selected_fields

        return {union_tag: final_map}
    elif isinstance(type_declaration_type, TObject):
        nested_type_declaration = type_declaration_type_params[0]
        value_as_map = cast(dict[str, object], value)

        final_map = {}
        for key, value in value_as_map.items():
            value_with_selected_fields = select_struct_fields(nested_type_declaration, value,
                                                              selected_struct_fields)

            final_map[key] = value_with_selected_fields

        return final_map
    elif isinstance(type_declaration_type, TArray):
        nested_type = type_declaration_type_params[0]
        value_as_list = cast(list[str], value)

        final_list = []
        for entry in value_as_list:
            value_with_selected_fields = select_struct_fields(
                nested_type, entry, selected_struct_fields)
            final_list.append(value_with_selected_fields)

        return final_list
    else:
        return value



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import msgpack
import json

from .Serialization import Serialization


class DefaultSerialization(Serialization):

    def to_json(self, telepact_message: object) -> bytes:
        return json.dumps(telepact_message).encode()

    def to_msgpack(self, telepact_message: object) -> bytes:
        return msgpack.dumps(telepact_message)

    def from_json(self, bytes_: bytes) -> object:
        return json.loads(bytes_)

    def from_msgpack(self, bytes_: bytes) -> object:
        return msgpack.loads(bytes_, strict_map_key=False)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast


if TYPE_CHECKING:
    from ...RandomGenerator import RandomGenerator
    from ..types.TFieldDeclaration import TFieldDeclaration
    from ..types.TTypeDeclaration import TTypeDeclaration
    from ...internal.generation.GenerateContext import GenerateContext


def generate_random_struct(blueprint_value: object, use_blueprint_value: bool,
                           reference_struct: dict[str, 'TFieldDeclaration'], ctx: 'GenerateContext') -> object:
    sorted_reference_struct = sorted(
        reference_struct.items(), key=lambda x: x[0])

    starting_struct = cast(
        dict[str, object], blueprint_value) if use_blueprint_value else {}

    obj = {}
    for field_name, field_declaration in sorted_reference_struct:
        this_use_blueprint_value = field_name in starting_struct
        type_declaration = field_declaration.type_declaration

        if this_use_blueprint_value:
            this_blueprint_value = starting_struct.get(field_name)
            value = type_declaration.generate_random_value(
                this_blueprint_value, this_use_blueprint_value, ctx)
        else:
            if not field_declaration.optional:
                if not ctx.always_include_required_fields and ctx.random_generator.next_boolean():
                    continue
                value = type_declaration.generate_random_value(
                    None, False, ctx)
            else:
                if not ctx.include_optional_fields or (ctx.randomize_optional_fields and ctx.random_generator.next_boolean()):
                    continue
                value = type_declaration.generate_random_value(
                    None, False, ctx)

        obj[field_name] = value

    return obj



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

from ...internal.binary.BinaryEncoder import BinaryEncoder

if TYPE_CHECKING:
    from .BinaryEncodingCache import BinaryEncodingCache

class ClientBinaryEncoder(BinaryEncoder):

    def __init__(self, binary_encoding_cache: 'BinaryEncodingCache') -> None:
        from .ClientBinaryStrategy import ClientBinaryStrategy
        self.binary_encoding_cache = binary_encoding_cache
        self.binaryChecksumStrategy = ClientBinaryStrategy(binary_encoding_cache)

    def encode(self, message: list[object]) -> list[object]:
        from ...internal.binary.ClientBinaryEncode import client_binary_encode
        return client_binary_encode(message, self.binary_encoding_cache, self.binaryChecksumStrategy)

    def decode(self, message: list[object]) -> list[object]:
        from ...internal.binary.ClientBinaryDecode import client_binary_decode
        return client_binary_decode(message, self.binary_encoding_cache, self.binaryChecksumStrategy)


#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
from msgpack import ExtType

from ...internal.binary.BinaryPackNode import BinaryPackNode


UNDEFINED_BYTE = 18


def pack_map(m: dict[object, object], header: list[object], key_index_map: dict[int, 'BinaryPackNode']) -> list[object]:
    from ...internal.binary.CannotPack import CannotPack
    from ...internal.binary.Pack import pack

    row: list[object] = []
    for key, value in m.items():
        if isinstance(key, str):
            raise CannotPack()

        key = cast(int, key)
        key_index = key_index_map.get(key)

        if key_index is None:
            final_key_index = BinaryPackNode(len(header) - 1, {})

            if isinstance(value, dict):
                header.append([key])
            else:
                header.append(key)

            key_index_map[key] = final_key_index
        else:
            final_key_index = key_index

        key_index_value = final_key_index.value
        key_index_nested = final_key_index.nested

        packed_value: object

        if isinstance(value, dict):
            try:
                nested_header = header[key_index_value + 1]
                if not isinstance(nested_header, list):
                    raise TypeError()
            except (IndexError, TypeError):
                raise CannotPack()

            packed_value = pack_map(value, nested_header, key_index_nested)
        else:
            if isinstance(header[key_index_value + 1], list):
                raise CannotPack()

            packed_value = pack(value)

        while len(row) < key_index_value:
            row.append(ExtType(UNDEFINED_BYTE, b''))

        if len(row) == key_index_value:
            row.append(packed_value)
        else:
            row[key_index_value] = packed_value

    return row



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .internal.schema.SchemaParseFailure import SchemaParseFailure


class TelepactSchemaParseError(Exception):
    """
    Indicates failure to parse a telepact Schema.
    """

    def __init__(self, schema_parse_failures: list['SchemaParseFailure'], telepact_schema_document_names_to_json: dict[str, str]):
        from .internal.schema.MapSchemaParseFailuresToPseudoJson import map_schema_parse_failures_to_pseudo_json
        super().__init__(str(map_schema_parse_failures_to_pseudo_json(
            schema_parse_failures, telepact_schema_document_names_to_json)))
        self.schema_parse_failures = schema_parse_failures
        self.schema_parse_failures_pseudo_json = map_schema_parse_failures_to_pseudo_json(
            schema_parse_failures, telepact_schema_document_names_to_json)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from collections import OrderedDict


def find_schema_key(document_name: str, definition: dict[str, object], index: int, document_names_to_json: dict[str, str]) -> str:
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ...internal.schema.SchemaParseFailure import SchemaParseFailure
    import re

    regex = "^(((fn|errors|headers|info)|((struct|union|_ext)(<[0-2]>)?))\\..*)"
    matches = []

    keys = sorted(list(definition.keys()))

    for e in keys:
        if re.match(regex, e):
            matches.append(e)

    if len(matches) == 1:
        return matches[0]
    else:
        parse_failure = SchemaParseFailure(document_name, [index],
                                           "ObjectKeyRegexMatchCountUnexpected",
                                           {"regex": regex, "actual": len(
                                               matches), "expected": 1, "keys": keys})
        raise TelepactSchemaParseError([parse_failure], document_names_to_json)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
from ...internal.schema.DerivePossibleSelects import derive_possible_select
from ...internal.schema.GetOrParseType import get_or_parse_type
from ...internal.schema.SchemaParseFailure import SchemaParseFailure
from ..types.TSelect import TSelect

if TYPE_CHECKING:
    from ...internal.schema.ParseContext import ParseContext
    from ..types.TType import TType
    from ..types.TUnion import TUnion


def parse_function_result_type(path: list[object], function_definition_as_parsed_json: dict[str, object],
                        schema_key: str,
                        ctx: 'ParseContext') -> 'TUnion':
    from ...internal.schema.ParseStructType import parse_struct_type
    from ...internal.schema.ParseUnionType import parse_union_type
    from ...internal.schema.ParseUnionType import parse_union_type
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ...internal.schema.SchemaParseFailure import SchemaParseFailure
    from ..types.TUnion import TUnion

    parse_failures = []

    result_schema_key = "->"

    result_type = None
    if result_schema_key not in function_definition_as_parsed_json:
        parse_failures.append(SchemaParseFailure(
            ctx.document_name, path, "RequiredObjectKeyMissing", {'key': result_schema_key}))
    else:
        try:
            result_type = parse_union_type(path, function_definition_as_parsed_json,
                                           result_schema_key, list(
                                               function_definition_as_parsed_json.keys()),
                                           ["Ok_"], ctx)
        except TelepactSchemaParseError as e:
            parse_failures.extend(e.schema_parse_failures)


    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, ctx.telepact_schema_document_names_to_json)

    fn_select_type = derive_possible_select(
        schema_key, cast(TUnion, result_type))
    select_type = cast(TSelect, get_or_parse_type([], '_ext.Select_', ctx))
    select_type.possible_selects[schema_key] = fn_select_type

    return cast(TUnion, result_type)

def parse_function_errors_regex(path: list[object], function_definition_as_parsed_json: dict[str, object],
                        schema_key: str,
                        ctx: 'ParseContext') -> str:
    from ...internal.schema.GetTypeUnexpectedParseFailure import get_type_unexpected_parse_failure
    from ...TelepactSchemaParseError import TelepactSchemaParseError

    parse_failures = []

    errors_regex_key = "_errors"

    regex_path = path + [errors_regex_key]

    errors_regex: str | None = None
    if errors_regex_key in function_definition_as_parsed_json and not schema_key.endswith("_"):
        parse_failures.append(SchemaParseFailure(
            ctx.document_name, regex_path, "ObjectKeyDisallowed", {}))
    else:
        errors_regex_init = function_definition_as_parsed_json.get(
            errors_regex_key, "^errors\\..*$")

        if not isinstance(errors_regex_init, str):
            this_parse_failures = get_type_unexpected_parse_failure(
                ctx.document_name, regex_path, errors_regex_init, "String")
            parse_failures.extend(this_parse_failures)
        else:
            errors_regex = errors_regex_init

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, ctx.telepact_schema_document_names_to_json)

    return cast(str, errors_regex)


#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import re
from typing import TYPE_CHECKING, cast
from ...TelepactSchemaParseError import TelepactSchemaParseError
from ...internal.schema.SchemaParseFailure import SchemaParseFailure

if TYPE_CHECKING:
    from ...internal.schema.ParseContext import ParseContext
    from ..types.TType import TType


def get_or_parse_type(path: list[object], type_name: str, ctx: 'ParseContext') -> 'TType':
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ..types.TObject import TObject
    from ..types.TArray import TArray
    from ..types.TBoolean import TBoolean
    from ..types.TInteger import TInteger
    from ..types.TNumber import TNumber
    from ..types.TObject import TObject
    from ..types.TString import TString
    from ..types.TBytes import TBytes
    from ..types.TMockCall import TMockCall
    from ..types.TMockStub import TMockStub
    from ..types.TSelect import TSelect
    from ..types.TUnion import TUnion
    from ..types.TAny import TAny
    from ...internal.schema.ParseFunctionType import parse_function_result_type
    from ...internal.schema.ParseFunctionType import parse_function_errors_regex
    from ...internal.schema.ParseStructType import parse_struct_type
    from ...internal.schema.ParseUnionType import parse_union_type

    if type_name in ctx.failed_types:
        raise TelepactSchemaParseError(
            [], ctx.telepact_schema_document_names_to_json)

    existing_type = ctx.parsed_types.get(type_name)
    if existing_type is not None:
        return existing_type

    regex_string = r"^(boolean|integer|number|string|any|bytes)|((fn|(union|struct|_ext))\.([a-zA-Z_]\w*))$"
    regex = re.compile(regex_string)

    matcher = regex.match(type_name)
    if not matcher:
        print(f"Type name '{type_name}' does not match expected regex: {regex_string}")
        raise TelepactSchemaParseError(
            [SchemaParseFailure(ctx.document_name, path, "StringRegexMatchFailed", {
                                "regex": regex_string})],
            ctx.telepact_schema_document_names_to_json)

    standard_type_name = matcher.group(1)
    if standard_type_name is not None:
        return {
            "boolean": TBoolean(),
            "integer": TInteger(),
            "number": TNumber(),
            "string": TString(),
            "bytes": TBytes()
        }.get(standard_type_name, TAny())

    custom_type_name = matcher.group(2)
    this_index = ctx.schema_keys_to_index.get(custom_type_name)
    this_document_name = cast(
        str, ctx.schema_keys_to_document_name.get(custom_type_name))
    if this_index is None:
        raise TelepactSchemaParseError(
            [SchemaParseFailure(ctx.document_name, path, "TypeUnknown", {
                                "name": custom_type_name})],
            ctx.telepact_schema_document_names_to_json)

    telepact_schema_pseudo_json = cast(
        list[object], ctx.telepact_schema_document_names_to_pseudo_json.get(this_document_name))
    definition = cast(
        dict[str, object], telepact_schema_pseudo_json[this_index])

    type: 'TType'
    try:
        this_path: list[object] = [this_index]
        if custom_type_name.startswith("struct"):
            type = parse_struct_type(this_path, definition, custom_type_name, [],
                                     ctx.copy(document_name=this_document_name))
            ctx.parsed_types[custom_type_name] = type
        elif custom_type_name.startswith("union"):
            type = parse_union_type(this_path, definition, custom_type_name, [], [],
                                    ctx.copy(document_name=this_document_name))
            
            ctx.parsed_types[custom_type_name] = type
        elif custom_type_name.startswith("fn"):
            arg_type = parse_struct_type(path, definition,
                                        custom_type_name, ["->", "_errors"],
                                        ctx.copy(document_name=this_document_name))
            type = TUnion(custom_type_name, {custom_type_name: arg_type}, {
                            custom_type_name: 0})

            ctx.parsed_types[custom_type_name] = type

            result_type = parse_function_result_type(
                this_path,
                definition,
                custom_type_name,
                ctx.copy(document_name=this_document_name)
            )

            ctx.parsed_types[custom_type_name + '.->'] = result_type

            errors_regex = parse_function_errors_regex(
                this_path,
                definition,
                custom_type_name,
                ctx.copy(document_name=this_document_name)
            )

            ctx.fn_error_regexes[custom_type_name] = errors_regex

        else:
            possible_type_extension = {
                '_ext.Select_': TSelect(),
                '_ext.Call_': TMockCall(ctx.parsed_types),
                '_ext.Stub_': TMockStub(ctx.parsed_types),
            }.get(custom_type_name)

            if not possible_type_extension:
                raise TelepactSchemaParseError([
                    SchemaParseFailure(
                        ctx.document_name,
                        [this_index],
                        'TypeExtensionImplementationMissing',
                        {'name': custom_type_name}
                    ),
                ], ctx.telepact_schema_document_names_to_json)

            type = possible_type_extension

            ctx.parsed_types[custom_type_name] = type


        return type
    except TelepactSchemaParseError as e:
        ctx.all_parse_failures.extend(e.schema_parse_failures)
        ctx.failed_types.add(custom_type_name)
        raise TelepactSchemaParseError(
            [], ctx.telepact_schema_document_names_to_json)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING
from ...internal.validation.ValidationFailure import ValidationFailure

if TYPE_CHECKING:
    from ..types.TFieldDeclaration import TFieldDeclaration


def validate_headers(headers: dict[str, object], parsed_request_headers: dict[str, 'TFieldDeclaration'], function_name: str) -> list['ValidationFailure']:
    from ...internal.validation.ValidateContext import ValidateContext

    validation_failures = []

    for header, header_value in headers.items():
        if not header.startswith("@"):
            validation_failures.append(ValidationFailure([header], "RequiredObjectKeyPrefixMissing", {"prefix": "@"}))

        field = parsed_request_headers.get(header)
        if field:
            this_validation_failures = field.type_declaration.validate(
                header_value, ValidateContext(None, function_name))
            this_validation_failures_path = [
                ValidationFailure([header] + e.path, e.reason, e.data)
                for e in this_validation_failures
            ]
            validation_failures.extend(this_validation_failures_path)

    return validation_failures



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import List, Dict, Tuple, Set, Union
from typing import TYPE_CHECKING

from ...internal.binary.BinaryEncoding import BinaryEncoding
from ...internal.binary.CreateChecksum import create_checksum

if TYPE_CHECKING:
    from ...TelepactSchema import TelepactSchema
    from ..types.TTypeDeclaration import TTypeDeclaration


def trace_type(type_declaration: 'TTypeDeclaration') -> list[str]:
    from ..types.TArray import TArray
    from ..types.TObject import TObject
    from ..types.TStruct import TStruct
    from ..types.TUnion import TUnion

    this_all_keys: list[str] = []

    if isinstance(type_declaration.type, TArray):
        these_keys2 = trace_type(type_declaration.type_parameters[0])
        this_all_keys.extend(these_keys2)
    elif isinstance(type_declaration.type, TObject):
        these_keys2 = trace_type(type_declaration.type_parameters[0])
        this_all_keys.extend(these_keys2)
    elif isinstance(type_declaration.type, TStruct):
        struct_fields = type_declaration.type.fields
        for struct_field_key, struct_field in struct_fields.items():
            this_all_keys.append(struct_field_key)
            more_keys = trace_type(struct_field.type_declaration)
            this_all_keys.extend(more_keys)
    elif isinstance(type_declaration.type, TUnion):
        union_tags = type_declaration.type.tags
        for tag_key, tag_value in union_tags.items():
            this_all_keys.append(tag_key)
            struct_fields = tag_value.fields
            for struct_field_key, struct_field in struct_fields.items():
                this_all_keys.append(struct_field_key)
                more_keys = trace_type(struct_field.type_declaration)
                this_all_keys.extend(more_keys)

    return this_all_keys


def construct_binary_encoding(telepact_schema: 'TelepactSchema') -> 'BinaryEncoding':
    from ..types.TUnion import TUnion

    all_keys: set[str] = set()

    for key, value in telepact_schema.parsed.items():

        if key.endswith('.->') and isinstance(value, TUnion):
            result = value.tags['Ok_']
            all_keys.add('Ok_')
            for field_key, field in result.fields.items():
                all_keys.add(field_key)
                keys = trace_type(field.type_declaration)
                for key in keys:
                    all_keys.add(key)

        elif key.startswith('fn.') and isinstance(value, TUnion):
            all_keys.add(key)
            args = value.tags[key]
            for field_key, field in args.fields.items():
                all_keys.add(field_key)
                keys = trace_type(field.type_declaration)
                for key in keys:
                    all_keys.add(key)

    sorted_all_keys = sorted(all_keys)

    binary_encoding = {key: i for i, key in enumerate(sorted_all_keys)}
    final_string = "\n".join(sorted_all_keys)
    checksum = create_checksum(final_string)
    return BinaryEncoding(binary_encoding, checksum)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Callable, TYPE_CHECKING, cast, Awaitable

from ..internal.binary.ServerBase64Decode import server_base64_decode

from ..Message import Message
from ..TelepactError import TelepactError
from .GetApiDefinitionsWithExamples import get_api_definitions_with_examples
from .types.TTypeDeclaration import TTypeDeclaration

if TYPE_CHECKING:
    from ..internal.validation.ValidationFailure import ValidationFailure
    from .types.TType import TType
    from ..TelepactSchema import TelepactSchema


async def handle_message(
    request_message: 'Message',
    override_headers: dict[str, object],
    telepact_schema: 'TelepactSchema',
    handler: Callable[['Message'], Awaitable['Message']],
    on_error: Callable[[Exception], None],
) -> 'Message':
    from ..internal.SelectStructFields import select_struct_fields
    from ..internal.validation.GetInvalidErrorMessage import get_invalid_error_message
    from ..internal.validation.MapValidationFailuresToInvalidFieldCases import map_validation_failures_to_invalid_field_cases
    from ..internal.validation.ValidateHeaders import validate_headers
    from ..internal.validation.ValidateResult import validate_result
    from .types.TUnion import TUnion
    from ..internal.validation.ValidateContext import ValidateContext

    response_headers: dict[str, object] = {}
    request_headers: dict[str, object] = request_message.headers
    request_body: dict[str, object] = request_message.body
    parsed_telepact_schema: dict[str, TType] = telepact_schema.parsed
    request_entry: tuple[str, object] = next(iter(request_body.items()))

    request_headers.update(override_headers)

    request_target_init = request_entry[0]
    request_payload = cast(
        dict[str, object], request_entry[1])

    unknown_target: str | None
    request_target: str
    if request_target_init not in parsed_telepact_schema:
        unknown_target = request_target_init
        request_target = "fn.ping_"
    else:
        unknown_target = None
        request_target = request_target_init

    function_name = request_target
    call_type = cast(TUnion, parsed_telepact_schema[request_target])
    result_union_type = cast(TUnion, parsed_telepact_schema[request_target + '.->'])

    call_id = request_headers.get("@id_")
    if call_id is not None:
        response_headers["@id_"] = call_id

    if "_parseFailures" in request_headers:
        parse_failures = cast(list[object], request_headers["_parseFailures"])
        new_error_result: dict[str, object] = {
            "ErrorParseFailure_": {"reasons": parse_failures}
        }

        validate_result(result_union_type, new_error_result)

        return Message(response_headers, new_error_result)

    request_header_validation_failures: list[ValidationFailure] = validate_headers(
        request_headers, telepact_schema.parsed_request_headers, function_name
    )
    if request_header_validation_failures:
        return get_invalid_error_message(
            "ErrorInvalidRequestHeaders_",
            request_header_validation_failures,
            result_union_type,
            response_headers,
        )

    if "@bin_" in request_headers:
        client_known_binary_checksums = cast(
            list[object], request_headers["@bin_"])

        response_headers["@binary_"] = True
        response_headers["@clientKnownBinaryChecksums_"] = client_known_binary_checksums

        if "@pac_" in request_headers:
            response_headers["@pac_"] = request_headers["@pac_"]

    select_struct_fields_header: dict[str, object] | None = cast(dict[str, object] | None, request_headers.get(
        "@select_"
    ))

    if unknown_target is not None:
        new_error_result = {
            "ErrorInvalidRequestBody_": {
                "cases": [
                    {
                        "path": [unknown_target],
                        "reason": {"FunctionUnknown": {}},
                    }
                ]
            }
        }

        validate_result(result_union_type, new_error_result)
        return Message(response_headers, new_error_result)

    function_type_call: TUnion = call_type

    call_validate_ctx = ValidateContext(None, function_name, coerce_base64=False)

    call_validation_failures: list[ValidationFailure] = function_type_call.validate(
        request_body, [], call_validate_ctx)
    if call_validation_failures:
        return get_invalid_error_message(
            "ErrorInvalidRequestBody_",
            call_validation_failures,
            result_union_type,
            response_headers,
        )
    
    if len(call_validate_ctx.bytes_coercions) > 0:
        server_base64_decode(request_body, call_validate_ctx.bytes_coercions)

    unsafe_response_enabled = cast(bool, request_headers.get("@unsafe_", False))

    call_message: Message = Message(
        request_headers, {request_target: request_payload})

    result_message: Message
    if function_name == "fn.ping_":
        result_message = Message({}, {"Ok_": {}})
    elif function_name == "fn.api_":
        include_internal = isinstance(request_payload, dict) and request_payload.get("includeInternal!") is True
        include_examples = isinstance(request_payload, dict) and request_payload.get("includeExamples!") is True
        api_definitions = get_api_definitions_with_examples(
            telepact_schema,
            include_internal,
        ) if include_examples else (
            telepact_schema.full if include_internal else telepact_schema.original
        )
        result_message = Message({}, {"Ok_": {"api": api_definitions}})
    else:
        try:
            result_message = await handler(call_message)
        except Exception as e:
            try:
                on_error(
                    TelepactError(
                        f"telepact handler failed while handling {function_name}",
                        kind="handler",
                        cause=e,
                    )
                )
            except Exception:
                pass
            return Message(response_headers, {"ErrorUnknown_": {}})

    result_union: dict[str, object] = result_message.body

    result_message.headers.update(response_headers)
    final_response_headers: dict[str, object] = result_message.headers

    skip_result_validation: bool = unsafe_response_enabled

    coerce_base64 = final_response_headers.get("@binary_", False) == False
    result_validate_ctx = ValidateContext(select_struct_fields_header, function_name, coerce_base64=coerce_base64)
    result_validation_failures: list[ValidationFailure] = result_union_type.validate(
        result_union, [], result_validate_ctx)
    
    if result_validation_failures and not skip_result_validation:
        try:
            on_error(
                TelepactError(
                    f"telepact response validation failed for {function_name}: {map_validation_failures_to_invalid_field_cases(result_validation_failures)}",
                    kind="validation",
                )
            )
        except Exception:
            pass
        res = get_invalid_error_message(
            "ErrorInvalidResponseBody_",
            result_validation_failures,
            result_union_type,
            final_response_headers,
        )
        return res
    
    if result_validate_ctx.base64_coercions:
        final_response_headers["@base64_"] = result_validate_ctx.base64_coercions

    if result_validate_ctx.bytes_coercions:
        server_base64_decode(result_union, result_validate_ctx.bytes_coercions)
    
    response_header_validation_failures: list[ValidationFailure] = validate_headers(
        final_response_headers, telepact_schema.parsed_response_headers, function_name
    )
    if response_header_validation_failures:
        try:
            on_error(
                TelepactError(
                    f"telepact response header validation failed for {function_name}: {map_validation_failures_to_invalid_field_cases(response_header_validation_failures)}",
                    kind="validation",
                )
            )
        except Exception:
            pass
        return get_invalid_error_message(
            "ErrorInvalidResponseHeaders_",
            response_header_validation_failures,
            result_union_type,
            response_headers,
        )

    final_result_union: dict[str, object]
    if select_struct_fields_header is not None:
        final_result_union = cast(dict[str, object], select_struct_fields(
            TTypeDeclaration(result_union_type, False, []),
            result_union,
            select_struct_fields_header,
        ))
    else:
        final_result_union = result_union

    return Message(final_response_headers, final_result_union)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
from ...internal.schema.SchemaParseFailure import SchemaParseFailure
from ..types.TStruct import TStruct

if TYPE_CHECKING:
    from ...internal.schema.ParseContext import ParseContext
    from ..types.TType import TType


def parse_struct_type(path: list[object], struct_definition_as_pseudo_json: dict[str, object],
                      schema_key: str, ignore_keys: list[str],
                      ctx: 'ParseContext') -> 'TStruct':
    from ...internal.schema.GetTypeUnexpectedParseFailure import get_type_unexpected_parse_failure
    from ...internal.schema.ParseStructFields import parse_struct_fields
    from ...TelepactSchemaParseError import TelepactSchemaParseError

    parse_failures = []
    other_keys = set(struct_definition_as_pseudo_json.keys())

    other_keys.discard(schema_key)
    other_keys.discard("///")
    other_keys.discard("_ignoreIfDuplicate")
    for ignore_key in ignore_keys:
        other_keys.discard(ignore_key)

    if other_keys:
        for k in other_keys:
            loop_path = path + [k]
            parse_failures.append(SchemaParseFailure(
                ctx.document_name, loop_path, "ObjectKeyDisallowed", {}))

    this_path = path + [schema_key]
    def_init = cast(dict[str, object],
                    struct_definition_as_pseudo_json.get(schema_key))

    definition = None
    if not isinstance(def_init, dict):
        branch_parse_failures = get_type_unexpected_parse_failure(
            ctx.document_name, this_path, def_init, "Object")
        parse_failures.extend(branch_parse_failures)
    else:
        definition = def_init

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, ctx.telepact_schema_document_names_to_json)

    fields = parse_struct_fields(this_path, definition, False, ctx)

    return TStruct(schema_key, fields)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import os
from typing import Dict

from ...internal.schema.SchemaParseFailure import SchemaParseFailure
from ...TelepactSchemaParseError import TelepactSchemaParseError
from .DocumentLocators import SchemaDocumentMap
from .ParseTelepactYaml import parse_telepact_yaml


def get_schema_file_map(directory: str) -> Dict[str, str]:
    final_json_documents: SchemaDocumentMap = SchemaDocumentMap()

    schema_parse_failures = []

    try:
        paths = [
            os.path.join(directory, entry)
            for entry in sorted(os.listdir(directory))
        ]

        for path in paths:
            relative_path = os.path.relpath(path, directory)

            # Check if directory
            # If it is, add a SchemaParseFailure for DirectoryDisallowed
            if os.path.isdir(path):
                schema_parse_failures.append(SchemaParseFailure(relative_path, [], "DirectoryDisallowed", {}))
                final_json_documents[relative_path] = "[]"
                continue    

            with open(path, 'r') as file:
                content = file.read()
            if relative_path.endswith('.telepact.json'):
                final_json_documents[relative_path] = content
            elif relative_path.endswith('.telepact.yaml'):
                try:
                    canonical_json, locator = parse_telepact_yaml(content)
                    final_json_documents[relative_path] = canonical_json
                    if locator is not None:
                        final_json_documents.document_locators[relative_path] = locator
                except ValueError:
                    final_json_documents[relative_path] = '[]'
                    schema_parse_failures.append(SchemaParseFailure(relative_path, [], "JsonInvalid", {}))
            else:
                final_json_documents[relative_path] = content
                schema_parse_failures.append(SchemaParseFailure(relative_path, [], "FileNamePatternInvalid", {"expected": "*.telepact.json|*.telepact.yaml"}))
    except IOError as e:
        raise RuntimeError(e)

    if schema_parse_failures:
        raise TelepactSchemaParseError(schema_parse_failures, final_json_documents)

    return final_json_documents



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...internal.binary.BinaryEncoding import BinaryEncoding


def decode_keys(given: object, binary_encoder: 'BinaryEncoding') -> object:
    from ...internal.binary.BinaryEncodingMissing import BinaryEncodingMissing

    if isinstance(given, dict):
        new_dict: dict[str, object] = {}

        for key, value in given.items():
            if isinstance(key, str):
                new_key = key
            else:
                possible_new_key = binary_encoder.decode_map.get(key)

                if possible_new_key is None:
                    raise BinaryEncodingMissing(key)

                new_key = possible_new_key

            encoded_value = decode_keys(value, binary_encoder)
            new_dict[new_key] = encoded_value

        return new_dict
    elif isinstance(given, list):
        return [decode_keys(item, binary_encoder) for item in given]
    else:
        return given



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from ...internal.validation.ValidateContext import ValidateContext
    from ...internal.validation.ValidationFailure import ValidationFailure


def validate_select(given_obj: object, possible_fn_selects: dict[str, object], ctx: 'ValidateContext') -> list['ValidationFailure']:
    from ...internal.validation.GetTypeUnexpectedValidationFailure import get_type_unexpected_validation_failure

    if not isinstance(given_obj, dict):
        return get_type_unexpected_validation_failure([], given_obj, 'Object')

    fn_scope = cast(str, ctx.fn)

    possible_select = possible_fn_selects[fn_scope]

    return is_sub_select([], given_obj, possible_select)


def is_sub_select(path: list[object], given_obj: object, possible_select_section: object) -> list['ValidationFailure']:
    from ...internal.validation.GetTypeUnexpectedValidationFailure import get_type_unexpected_validation_failure
    from ...internal.validation.ValidationFailure import ValidationFailure

    if isinstance(possible_select_section, list):
        if not isinstance(given_obj, list):
            return get_type_unexpected_validation_failure(path, given_obj, 'Array')

        validation_failures = []

        for index, element in enumerate(given_obj):
            if element not in possible_select_section:
                validation_failures.append(ValidationFailure(
                    path + [index], 'ArrayElementDisallowed', {}))

        return validation_failures

    elif isinstance(possible_select_section, dict):
        if not isinstance(given_obj, dict):
            return get_type_unexpected_validation_failure(path, given_obj, 'Object')

        validation_failures = []

        for key, value in given_obj.items():
            if key in possible_select_section:
                inner_failures = is_sub_select(
                    path + [key], value, possible_select_section[key])
                validation_failures.extend(inner_failures)
            else:
                validation_failures.append(ValidationFailure(
                    path + [key], 'ObjectKeyDisallowed', {}))

        return validation_failures

    return []



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

class SerializationError(Exception):
    def __init__(
        self,
        message: str = "telepact serialization failed",
        *,
        context: str | None = None,
        cause: BaseException | None = None,
    ) -> None:
        if context:
            message = f"{message} while {context}"
        if cause is not None:
            message = f"{message}: {cause}"
        super().__init__(message)
        self.context = context
        self.cause = cause



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING
import json
import re

if TYPE_CHECKING:
    from ...TelepactSchema import TelepactSchema


def create_telepact_schema_from_file_json_map(json_documents: dict[str, str]) -> 'TelepactSchema':
    from .ParseTelepactSchema import parse_telepact_schema
    from .GetInternalTelepactJson import get_internal_telepact_json
    from .GetAuthTelepactJson import get_auth_telepact_json
    from .DocumentLocators import copy_document_locators

    final_json_documents = copy_document_locators(json_documents, dict(json_documents))
    internal_json = get_internal_telepact_json()
    if not _has_bundled_definitions(json_documents, "internal_", internal_json):
        final_json_documents["internal_"] = internal_json

    # Determine if we need to add the auth schema
    auth_json = get_auth_telepact_json()
    for json in json_documents.values():
        regex = re.compile(r'"struct\.Auth_"\s*:')
        matcher = regex.search(json)
        if matcher:
            if not _has_bundled_definitions(json_documents, "auth_", auth_json):
                final_json_documents["auth_"] = auth_json
            break

    return parse_telepact_schema(final_json_documents)


def _has_bundled_definitions(json_documents: dict[str, str], bundled_document_name: str, bundled_json: str) -> bool:
    bundled_keys = _collect_schema_keys({bundled_document_name: bundled_json})
    if bundled_keys is None:
        return False

    provided_keys = _collect_schema_keys(json_documents)
    if provided_keys is None:
        return False

    return bundled_keys.issubset(provided_keys)


def _collect_schema_keys(json_documents: dict[str, str]) -> set[str] | None:
    from .FindSchemaKey import find_schema_key

    schema_keys: set[str] = set()

    for document_name, document_json in json_documents.items():
        try:
            pseudo_json = json.loads(document_json)
        except json.JSONDecodeError:
            return None

        if not isinstance(pseudo_json, list):
            return None

        for index, definition in enumerate(pseudo_json):
            if not isinstance(definition, dict):
                continue

            try:
                schema_keys.add(find_schema_key(document_name, definition, index, json_documents))
            except Exception:
                return None

    return schema_keys



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Callable, TYPE_CHECKING

from ..Message import Message
from ..internal.binary.BinaryEncoderUnavailableError import BinaryEncoderUnavailableError
from ..internal.binary.BinaryEncodingMissing import BinaryEncodingMissing
from ..internal.validation.InvalidMessage import InvalidMessage
from ..internal.validation.InvalidMessageBody import InvalidMessageBody
from ..TelepactError import TelepactError

if TYPE_CHECKING:
    from ..Serializer import Serializer
    from ..TelepactSchema import TelepactSchema


def parse_request_message(request_message_bytes: bytes, serializer: 'Serializer', telepact_schema: 'TelepactSchema',
                          on_error: Callable[[Exception], None]) -> 'Message':
    try:
        return serializer.deserialize(request_message_bytes)
    except Exception as e:
        on_error(
            TelepactError(
                "telepact request parsing failed while decoding the incoming message",
                kind="parse",
                cause=e,
            )
        )

        if isinstance(e, BinaryEncoderUnavailableError):
            reason = "IncompatibleBinaryEncoding"
        elif isinstance(e, BinaryEncodingMissing):
            reason = "BinaryDecodeFailure"
        elif isinstance(e, InvalidMessage):
            reason = "ExpectedJsonArrayOfTwoObjects"
        elif isinstance(e, InvalidMessageBody):
            reason = "ExpectedJsonArrayOfAnObjectAndAnObjectOfOneObject"
        else:
            reason = "ExpectedJsonArrayOfTwoObjects"

        return Message({"_parseFailures": [{reason: {}}]}, {"_unknown": {}})



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import cast, TYPE_CHECKING
from ...internal.binary.BinaryEncoding import BinaryEncoding
from ...internal.binary.BinaryEncoderUnavailableError import BinaryEncoderUnavailableError

if TYPE_CHECKING:
    from .BinaryEncodingCache import BinaryEncodingCache
    from .ClientBinaryStrategy import ClientBinaryStrategy


def client_binary_encode(message: list[object], binary_encoding_cache: 'BinaryEncodingCache',
                         binary_checksum_strategy: 'ClientBinaryStrategy') -> list[object]:
    from ...internal.binary.EncodeBody import encode_body
    from ...internal.binary.PackBody import pack_body

    headers = cast(dict[str, object], message[0])
    message_body = cast(dict[str, object], message[1])
    force_send_json = headers.pop("_forceSendJson", None)

    checksums = binary_checksum_strategy.get_current_checksums()
    headers["@bin_"] = checksums

    if force_send_json == True:
        raise BinaryEncoderUnavailableError()
    
    if len(checksums) > 1:
        raise BinaryEncoderUnavailableError()
    
    binary_encoding = binary_encoding_cache.get(checksums[0]) if checksums else None

    if not binary_encoding:
        raise BinaryEncoderUnavailableError()

    encoded_message_body = encode_body(message_body, binary_encoding)

    final_encoded_message_body = pack_body(encoded_message_body) if headers.get(
        "@pac_") == True else encoded_message_body

    return [headers, final_encoded_message_body]



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
from ...internal.validation.ValidationFailure import ValidationFailure

if TYPE_CHECKING:
    from ...internal.validation.ValidateContext import ValidateContext
    from ..types.TStruct import TStruct
    from ..types.TTypeDeclaration import TTypeDeclaration


def validate_union_struct(
    union_struct: 'TStruct',
    union_tag: str,
    actual: dict[str, object],
    selected_tags: dict[str, object],
    ctx: 'ValidateContext'
) -> list['ValidationFailure']:
    selected_fields = cast(list[str], selected_tags.get(
        union_tag)) if selected_tags else None
    from ...internal.validation.ValidateStructFields import validate_struct_fields

    ctx.path.append(union_tag)

    result = validate_struct_fields(
        union_struct.fields,
        selected_fields,
        actual,
        ctx
    )

    ctx.path.pop()

    return result



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
from ...internal.schema.SchemaParseFailure import SchemaParseFailure
from ..types.TError import TError

if TYPE_CHECKING:
    from ...internal.schema.ParseContext import ParseContext


def parse_error_type(path: list[object], error_definition_as_parsed_json: dict[str, object],
                     schema_key: str,
                     ctx: 'ParseContext') -> 'TError':
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ...internal.schema.ParseUnionType import parse_union_type

    parse_failures = []

    other_keys = set(error_definition_as_parsed_json.keys())
    other_keys.discard(schema_key)
    other_keys.discard("///")

    if other_keys:
        for k in other_keys:
            loop_path = path + [k]
            parse_failures.append(SchemaParseFailure(
                ctx.document_name, cast(list[object], loop_path), "ObjectKeyDisallowed", {}))

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, ctx.telepact_schema_document_names_to_json)

    error = parse_union_type(path, error_definition_as_parsed_json, schema_key, [], [],
                             ctx)

    return TError(schema_key, error)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

class BinaryEncoding:
    def __init__(self, binary_encoding_map: dict[str, int], checksum: int) -> None:
        self.encode_map: dict[str, int] = binary_encoding_map
        self.decode_map: dict[int, str] = {
            v: k for k, v in binary_encoding_map.items()}
        self.checksum: int = checksum



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Callable, TYPE_CHECKING, Awaitable, NamedTuple

from .DefaultSerialization import DefaultSerialization
from .Serializer import Serializer
from .internal.binary.ServerBinaryEncoder import ServerBinaryEncoder
from .internal.binary.ServerBase64Encoder import ServerBase64Encoder

if TYPE_CHECKING:
    from .Message import Message
    from .TelepactSchema import TelepactSchema
    from .Response import Response

class Server:
    """
    A telepact Server.
    """
    class Options:
        """
        Options for the Server.
        """

        def __init__(self) -> None:
            self.on_error = lambda e: None
            self.on_request = lambda m: None
            self.on_response = lambda m: None
            self.auth_required = True
            self.serialization = DefaultSerialization()

    def __init__(self, telepact_schema: 'TelepactSchema', handler: Callable[['Message'], Awaitable['Message']], options: Options):
        """
        Create a server with the given telepact schema and handler.
        """
        from .internal.binary.ConstructBinaryEncoding import construct_binary_encoding

        self.handler = handler
        self.on_error = options.on_error
        self.on_request = options.on_request
        self.on_response = options.on_response

        self.telepact_schema = telepact_schema

        binary_encoding = construct_binary_encoding(self.telepact_schema)
        binary_encoder = ServerBinaryEncoder(binary_encoding)
        base64_encoder = ServerBase64Encoder()
        self.serializer = Serializer(options.serialization, binary_encoder, base64_encoder)

        if "struct.Auth_" not in self.telepact_schema.parsed and options.auth_required:
            raise RuntimeError(
                "Unauthenticated server. Either define a `struct.Auth_` in your schema or set `options.auth_required` to `false`."
            )

    async def process(self, request_message_bytes: bytes, override_headers: dict[str, object] = {}) -> 'Response':
        """
        Process a given telepact Request Message into a telepact Response Message.
        """
        from .internal.ProcessBytes import process_bytes

        return await process_bytes(request_message_bytes, override_headers, self.serializer, self.telepact_schema, self.on_error,
                                   self.on_request, self.on_response, self.handler)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast


if TYPE_CHECKING:
    from ...RandomGenerator import RandomGenerator
    from ..types.TStruct import TStruct
    from ..types.TTypeDeclaration import TTypeDeclaration
    from ...internal.generation.GenerateContext import GenerateContext


def generate_random_union(blueprint_value: object, use_blueprint_value: bool,
                          union_tags_reference: dict[str, 'TStruct'], ctx: 'GenerateContext') -> object:
    from ...internal.generation.GenerateRandomStruct import generate_random_struct

    # TODO: remove this comment.

    if not use_blueprint_value:
        sorted_union_tags_reference = sorted(
            union_tags_reference.items(), key=lambda x: x[0])
        random_index = ctx.random_generator.next_int_with_ceiling(
            len(sorted_union_tags_reference))
        union_tag, union_data = sorted_union_tags_reference[random_index]
        return {union_tag: generate_random_struct(None, False, union_data.fields, ctx)}
    else:
        starting_union = cast(dict[str, object], blueprint_value)
        union_tag, union_starting_struct = cast(
            tuple[str, dict[str, object]], next(iter(starting_union.items())))
        union_struct_type = union_tags_reference[union_tag]
        return {union_tag: generate_random_struct(union_starting_struct, True, union_struct_type.fields, ctx)}



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import json
from typing import Callable

import yaml


def _serialize_path(path: list[object]) -> str:
    return json.dumps(path, separators=(',', ':'))


def _coordinates_from_mark(mark: yaml.Mark | None) -> dict[str, object]:
    if mark is None:
        return {'row': 1, 'col': 1}
    return {'row': mark.line + 1, 'col': mark.column + 1}


def _build_locations_for_node(
    node: yaml.nodes.Node | None,
    path: list[object],
    locations: dict[str, dict[str, object]],
) -> None:
    if isinstance(node, yaml.nodes.SequenceNode):
        for index, item in enumerate(node.value):
            item_path = path + [index]
            locations[_serialize_path(item_path)] = _coordinates_from_mark(item.start_mark)
            _build_locations_for_node(item, item_path, locations)
        return

    if isinstance(node, yaml.nodes.MappingNode):
        for key_node, value_node in node.value:
            key = getattr(key_node, 'value', None)
            if not isinstance(key, str):
                continue

            key_path = path + [key]
            locations[_serialize_path(key_path)] = _coordinates_from_mark(key_node.start_mark)
            _build_locations_for_node(value_node, key_path, locations)


def _validate_no_duplicate_keys(node: yaml.nodes.Node | None) -> None:
    if isinstance(node, yaml.nodes.SequenceNode):
        for item in node.value:
            _validate_no_duplicate_keys(item)
        return

    if isinstance(node, yaml.nodes.MappingNode):
        seen_keys: set[str] = set()
        for key_node, value_node in node.value:
            key = str(getattr(key_node, 'value', ''))
            if key in seen_keys:
                raise ValueError('Duplicate YAML key')
            seen_keys.add(key)
            _validate_no_duplicate_keys(value_node)


def create_document_locator_from_yaml_text(text: str) -> Callable[[list[object]], dict[str, object]]:
    node = yaml.compose(text)
    if node is None:
        raise ValueError('YAML document cannot be empty')

    _validate_no_duplicate_keys(node)
    locations: dict[str, dict[str, object]] = {
        _serialize_path([]): _coordinates_from_mark(node.start_mark),
    }
    _build_locations_for_node(node, [], locations)

    return lambda path: locations.get(_serialize_path(path), {'row': 1, 'col': 1})



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import cast


def catch_error_collisions(telepact_schema_name_to_pseudo_json: dict[str, list[object]], error_keys: set[str], keys_to_index: dict[str, int], schema_keys_to_document_name: dict[str, str], document_names_to_json: dict[str, str]) -> None:
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ...internal.schema.SchemaParseFailure import SchemaParseFailure
    from ...internal.schema.DocumentLocators import resolve_document_coordinates

    parse_failures = []
    error_keys_list = list(error_keys)
    error_keys_list.sort(key=lambda k: (
        schema_keys_to_document_name[k], keys_to_index[k]))

    for i in range(len(error_keys_list)):
        for j in range(i + 1, len(error_keys_list)):
            def_key = error_keys_list[i]
            other_def_key = error_keys_list[j]

            index = keys_to_index[def_key]
            other_index = keys_to_index[other_def_key]

            document_name = schema_keys_to_document_name[def_key]
            other_document_name = schema_keys_to_document_name[other_def_key]

            telepact_schema_pseudo_json = telepact_schema_name_to_pseudo_json[document_name]
            other_telepact_schema_pseudo_json = telepact_schema_name_to_pseudo_json[other_document_name]

            def_ = cast(dict[str, object], telepact_schema_pseudo_json[index])
            other_def = cast(dict[str, object],
                             other_telepact_schema_pseudo_json[other_index])

            err_def = cast(list[object], def_[def_key])
            other_err_def = cast(list[object], other_def[other_def_key])

            for k in range(len(err_def)):
                this_err_def = cast(dict[str, object], err_def[k])
                this_err_def_keys = set(this_err_def.keys())
                this_err_def_keys.discard("///")

                for l in range(len(other_err_def)):
                    this_other_err_def = cast(
                        dict[str, object], other_err_def[l])
                    this_other_err_def_keys = set(this_other_err_def.keys())
                    this_other_err_def_keys.discard("///")

                    if this_err_def_keys == this_other_err_def_keys:
                        this_error_def_key = next(iter(this_err_def_keys))
                        this_other_error_def_key = next(
                            iter(this_other_err_def_keys))
                        final_this_path = [
                            index, def_key, k, this_error_def_key]
                        final_this_location_pseudo_json = resolve_document_coordinates(
                            final_this_path, document_name, document_names_to_json)
                        parse_failures.append(SchemaParseFailure(
                            other_document_name,
                            [other_index, other_def_key, l,
                                this_other_error_def_key],
                            "PathCollision",
                            {"document": document_name,
                             "path": final_this_path,
                             "location": final_this_location_pseudo_json}
                        ))

    if parse_failures:
        raise TelepactSchemaParseError(parse_failures, document_names_to_json)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...internal.binary.BinaryEncoding import BinaryEncoding


def encode_keys(given: object, binary_encoding: 'BinaryEncoding') -> object:
    if given is None:
        return given
    elif isinstance(given, dict):
        new_dict: dict[object, object] = {}

        for key, value in given.items():
            final_key = binary_encoding.encode_map.get(key, key)
            encoded_value = encode_keys(value, binary_encoding)

            new_dict[final_key] = encoded_value

        return new_dict
    elif isinstance(given, list):
        return [encode_keys(item, binary_encoding) for item in given]
    else:
        return given



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import re
from typing import TYPE_CHECKING

from ...internal.schema.SchemaParseFailure import SchemaParseFailure
from ..types.TTypeDeclaration import TTypeDeclaration

if TYPE_CHECKING:
    from ..types.TType import TType
    from ...internal.schema.ParseContext import ParseContext

def parse_type_declaration(
    path: list[object],
    type_declaration_object: object,
    ctx: 'ParseContext'
) -> 'TTypeDeclaration':
    from ...internal.types.TArray import TArray
    from ...internal.types.TObject import TObject
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ...internal.schema.GetOrParseType import get_or_parse_type
    from ...internal.schema.GetTypeUnexpectedParseFailure import get_type_unexpected_parse_failure

    if isinstance(type_declaration_object, str):
        root_type_string = type_declaration_object

        regex_string = r"^(.*?)(\?)?$"
        regex = re.compile(regex_string)

        matcher = regex.match(root_type_string)
        if not matcher:
            raise TelepactSchemaParseError(
                [SchemaParseFailure(ctx.document_name, path,
                                    "StringRegexMatchFailed", {"regex": regex_string})],
                ctx.telepact_schema_document_names_to_json
            )

        type_name = matcher.group(1)
        nullable = matcher.group(2) is not None

        t_type = get_or_parse_type(path, type_name, ctx)

        if t_type.get_type_parameter_count() != 0:
            raise TelepactSchemaParseError(
                [SchemaParseFailure(ctx.document_name, path,
                                    "ArrayLengthUnexpected",
                                    {"actual": 1, "expected": t_type.get_type_parameter_count() + 1})],
                ctx.telepact_schema_document_names_to_json
            )

        return TTypeDeclaration(t_type, nullable, [])

    elif isinstance(type_declaration_object, list):
        list_object = type_declaration_object

        if len(list_object) != 1:
            raise TelepactSchemaParseError(
                [SchemaParseFailure(ctx.document_name, path,
                                    "ArrayLengthUnexpected", {"actual": len(list_object), "expected": 1})],
                ctx.telepact_schema_document_names_to_json
            )

        element_type_declaration = list_object[0]
        new_path = path + [0]

        array_type = TArray()
        parsed_element_type = parse_type_declaration(new_path, element_type_declaration, ctx)

        return TTypeDeclaration(array_type, False, [parsed_element_type])

    elif isinstance(type_declaration_object, dict):
        map_object = type_declaration_object

        if len(map_object) != 1:
            raise TelepactSchemaParseError(
                [SchemaParseFailure(ctx.document_name, path,
                                    "ObjectSizeUnexpected", {"actual": len(map_object), "expected": 1})],
                ctx.telepact_schema_document_names_to_json
            )

        key, value = next(iter(map_object.items()))

        if not isinstance(key, str) or key != "string":
            key_path = path + [key]
            raise TelepactSchemaParseError(
                [
                    SchemaParseFailure(ctx.document_name, path, "RequiredObjectKeyMissing", {"key": "string"}),
                    SchemaParseFailure(ctx.document_name, key_path, "ObjectKeyDisallowed", {}),
                ],
                ctx.telepact_schema_document_names_to_json
            )

        new_path = path + [key]

        object_type = TObject()
        parsed_value_type = parse_type_declaration(new_path, value, ctx)

        return TTypeDeclaration(object_type, False, [parsed_value_type])

    else:
        failures = get_type_unexpected_parse_failure(
            ctx.document_name, path, type_declaration_object, "StringOrArrayOrObject"
        )
        raise TelepactSchemaParseError(failures, ctx.telepact_schema_document_names_to_json)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

from ...internal.validation.ValidationFailure import ValidationFailure

if TYPE_CHECKING:
    from ...internal.validation.ValidateContext import ValidateContext
    from ..types.TFieldDeclaration import TFieldDeclaration
    from ..types.TTypeDeclaration import TTypeDeclaration


def validate_struct_fields(fields: dict[str, 'TFieldDeclaration'],
                           selected_fields: list[str] | None,
                           actual_struct: dict[str, object],
                           ctx: 'ValidateContext') -> list['ValidationFailure']:
    validation_failures = []

    missing_fields = []
    for field_name, field_declaration in fields.items():
        is_optional = field_declaration.optional
        is_omitted_by_select = selected_fields is not None and field_name not in selected_fields
        if field_name not in actual_struct and not is_optional and not is_omitted_by_select:
            missing_fields.append(field_name)

    for missing_field in missing_fields:
        validation_failure = ValidationFailure(
            [], "RequiredObjectKeyMissing", {'key': missing_field})

        validation_failures.append(validation_failure)

    for field_name, field_value in actual_struct.items():
        reference_field = fields.get(field_name)
        if reference_field is None:
            validation_failure = ValidationFailure(
                [field_name], "ObjectKeyDisallowed", {})

            validation_failures.append(validation_failure)
            continue

        ref_field_type_declaration = reference_field.type_declaration

        ctx.path.append(field_name)

        nested_validation_failures = ref_field_type_declaration.validate(
            field_value, ctx)
        
        ctx.path.pop()
        
        nested_validation_failures_with_path = []
        for failure in nested_validation_failures:
            this_path = [field_name] + failure.path

            nested_validation_failures_with_path.append(
                ValidationFailure(this_path, failure.reason, failure.data))

        validation_failures.extend(nested_validation_failures_with_path)

    return validation_failures



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import re
from typing import TYPE_CHECKING
from ...internal.schema.SchemaParseFailure import SchemaParseFailure
from ..types.TFieldDeclaration import TFieldDeclaration

if TYPE_CHECKING:
    from ..types.TType import TType
    from ...internal.schema.ParseContext import ParseContext


def parse_field(path: list[object], field_declaration: str, type_declaration_value: object,
                is_header: bool,
                ctx: 'ParseContext') -> 'TFieldDeclaration':
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ...internal.schema.GetTypeUnexpectedParseFailure import get_type_unexpected_parse_failure
    from ...internal.schema.ParseTypeDeclaration import parse_type_declaration

    header_regex_string = r"^@[a-z][a-zA-Z0-9_]*$"
    regex_string = r"^([a-z][a-zA-Z0-9_]*)(!)?$"
    regex_to_use = header_regex_string if is_header else regex_string
    regex = re.compile(regex_to_use)

    matcher = regex.match(field_declaration)
    if not matcher:
        final_path = path + [field_declaration]
        raise TelepactSchemaParseError([SchemaParseFailure(ctx.document_name, final_path,
                                                          "KeyRegexMatchFailed",
                                                          {"regex": regex_to_use})], ctx.telepact_schema_document_names_to_json)

    field_name = matcher.group(0)
    optional = True if is_header else bool(matcher.group(2))

    this_path = path + [field_name]

    type_declaration = parse_type_declaration(this_path,
                                              type_declaration_value, ctx)

    return TFieldDeclaration(field_name, type_declaration, optional)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from .BuildDocumentLocatorFromYamlAst import create_document_locator_from_yaml_text


def create_path_document_yaml_coordinates_pseudo_json_locator(text: str):
    return create_document_locator_from_yaml_text(text)


def get_path_document_yaml_coordinates_pseudo_json(path: list[object], document: str) -> dict[str, object]:
    return create_path_document_yaml_coordinates_pseudo_json_locator(document)(path)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Callable


DocumentLocator = Callable[[list[object]], dict[str, object]]


class SchemaDocumentMap(dict[str, str]):
    document_locators: dict[str, DocumentLocator]

    def __init__(self, *args: object, **kwargs: object):
        super().__init__(*args, **kwargs)
        self.document_locators = {}


def ensure_schema_document_map(document_names_to_json: dict[str, str]) -> SchemaDocumentMap:
    if isinstance(document_names_to_json, SchemaDocumentMap):
        return document_names_to_json

    mapped = SchemaDocumentMap(document_names_to_json)
    locators = getattr(document_names_to_json, 'document_locators', None)
    if isinstance(locators, dict):
        mapped.document_locators = dict(locators)
    return mapped


def set_document_locators(document_names_to_json: dict[str, str], document_locators: dict[str, DocumentLocator]) -> None:
    mapped = ensure_schema_document_map(document_names_to_json)
    mapped.document_locators = dict(document_locators)


def get_document_locators(document_names_to_json: dict[str, str]) -> dict[str, DocumentLocator]:
    mapped = ensure_schema_document_map(document_names_to_json)
    return mapped.document_locators


def copy_document_locators(src: dict[str, str], dst: dict[str, str]) -> SchemaDocumentMap:
    mapped = ensure_schema_document_map(dst)
    mapped.document_locators = dict(getattr(src, 'document_locators', {}))
    return mapped


def resolve_document_coordinates(path: list[object], document_name: str, document_names_to_json: dict[str, str]) -> dict[str, object]:
    from .GetPathDocumentCoordinatesPseudoJson import get_path_document_coordinates_pseudo_json

    locator = getattr(document_names_to_json, 'document_locators', {}).get(document_name)
    if locator is not None:
        return locator(path)

    return get_path_document_coordinates_pseudo_json(path, document_names_to_json.get(document_name, '[]'))



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import cast, NamedTuple


class Message(NamedTuple):
    headers: dict[str, object]
    body: dict[str, object]

    def get_body_target(self) -> str:
        entry = next(iter(self.body.items()))
        return entry[0]

    def get_body_payload(self) -> dict[str, object]:
        entry = next(iter(self.body.items()))
        return cast(dict[str, object], entry[1])



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Any, TYPE_CHECKING

from ...internal.binary.BinaryEncodingCache import BinaryEncodingCache

if TYPE_CHECKING:
    from .BinaryEncoding import BinaryEncoding

class DefaultBinaryEncodingCache(BinaryEncodingCache):

    def __init__(self) -> None:
        self.recent_binary_encoders: dict[int, 'BinaryEncoding'] = {}

    def add(self, checksum: int, binary_encoding_map: dict[str, int]) -> None:
        from .BinaryEncoding import BinaryEncoding        
        binary_encoding = BinaryEncoding(binary_encoding_map, checksum)
        self.recent_binary_encoders[checksum] = binary_encoding

    def get(self, checksum: int) -> 'BinaryEncoding':
        return self.recent_binary_encoders[checksum]

    def remove(self, checksum: int) -> None:
        self.recent_binary_encoders.pop(checksum, None)


#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING
from ...TelepactError import TelepactError

if TYPE_CHECKING:
    from ..types.TUnion import TUnion


def validate_result(result_union_type: 'TUnion', error_result: object) -> None:
    from ...internal.validation.MapValidationFailuresToInvalidFieldCases import map_validation_failures_to_invalid_field_cases
    from ...internal.validation.ValidateContext import ValidateContext

    new_error_result_validation_failures = result_union_type.validate(
        error_result, [], ValidateContext(None, None)
    )
    if new_error_result_validation_failures:
        raise TelepactError(
            f"Failed internal telepact validation: {map_validation_failures_to_invalid_field_cases(new_error_result_validation_failures)}"
        )
